(* generated by dune *)

(** @canonical Re.Ast *)
module Ast = Re__Ast

(** @canonical Re.Automata *)
module Automata = Re__Automata

(** @canonical Re.Bit_vector *)
module Bit_vector = Re__Bit_vector

(** @canonical Re.Category *)
module Category = Re__Category

(** @canonical Re.Color_map *)
module Color_map = Re__Color_map

(** @canonical Re.Compile *)
module Compile = Re__Compile

(** @canonical Re.Core *)
module Core = Re__Core

(** @canonical Re.Cset *)
module Cset = Re__Cset

(** @canonical Re.Dense_map *)
module Dense_map = Re__Dense_map

(** @canonical Re.Dyn *)
module Dyn = Re__Dyn

(** @canonical Re.Emacs *)
module Emacs = Re__Emacs

(** @canonical Re.Fmt *)
module Fmt = Re__Fmt

(** @canonical Re.Glob *)
module Glob = Re__Glob

(** @canonical Re.Group *)
module Group = Re__Group

(** @canonical Re.Hash_set *)
module Hash_set = Re__Hash_set

(** @canonical Re.Import *)
module Import = Re__Import

(** @canonical Re.Mark_infos *)
module Mark_infos = Re__Mark_infos

(** @canonical Re.Parse_buffer *)
module Parse_buffer = Re__Parse_buffer

(** @canonical Re.Pcre *)
module Pcre = Re__Pcre

(** @canonical Re.Perl *)
module Perl = Re__Perl

(** @canonical Re.Pmark *)
module Pmark = Re__Pmark

(** @canonical Re.Posix *)
module Posix = Re__Posix

(** @canonical Re.Posix_class *)
module Posix_class = Re__Posix_class

(** @canonical Re.Replace *)
module Replace = Re__Replace

(** @canonical Re.Search *)
module Search = Re__Search

(** @canonical Re.Slice *)
module Slice = Re__Slice

(** @canonical Re.Str *)
module Str = Re__Str

(** @canonical Re.View *)
module View = Re__View

module Re__ = struct end
[@@deprecated "this module is shadowed"]
